==> mater-0.7.0-py3-none-any.whl: mater/_utils.py <==
"""
This module creates the external funcions used in the MATER class.
"""

from functools import wraps
from typing import Dict, List

import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame

# Functions definition


def _log_normal(time: DataFrame, life_time: DataFrame, standard_deviation: DataFrame, k: int) -> DataFrame:
    r"""
    Compute the log-normal probability density function.

    This function calculates the log-normal distribution given the lifetime mean value, standard deviation, and a set of time points. It is typically used to model time until failure or time until an event occurs, where the event times are assumed to follow a log-normal distribution.

    :param time: :math:`t` a DataFrame containing time points at which the log-normal distribution is evaluated.
    :param life_time: :math:`\mu_{l,o,t}` a DataFrame containing the lifetime mean values.
    :param standard_deviation: :math:`\sigma_{l,o,t}` a DataFrame containing the standard deviations of the lifetime values.
    :param k: A `time` integer.
    :type time: pd.DataFrame
    :type life_time: pd.DataFrame
    :type standard_deviation: pd.DataFrame
    :type k: int
    :return: :math:`d^{log}_{l,o,t}` a DataFrame containing the computed log-normal probability densities at each time point.
    :rtype: pd.DataFrame

    Calculation
    -----------
    The probability density function for a log-normal distribution is calculated as follows:

    .. math::
        d^{log}_{l,o,t}[t,\mu_{l,o,t},\sigma_{l,o,t},k] = \frac{1}{(t - k) \sigma_{l,o,t} \sqrt{2\pi}} \exp\left(-\frac{(\log(t - k) - \mu_{l,o,t})^2}{2\sigma_{l,o,t}^2}\right)
        :label: probability_density_function

    Notes
    -----
    Indices are the :ref:`variables dimensions <dimensions>`.
    """
    return np.exp(
        -1
        / 2
        * ((np.log(time.loc[:, k + 1 :] - k).subtract(life_time[k], axis=0)).div(standard_deviation[k], axis=0)) ** 2
    ) / ((time.loc[:, k + 1 :] - k).mul(standard_deviation[k], axis=0) * np.sqrt(2 * np.pi))


def _exponential_decay(time: DataFrame, life_time: DataFrame, k: int):
    r"""
    Calculate the exponential decay for given time points and decay lifetimes.

    :param time: :math:`t` is a `Time` DataFrame fot `time` :math:`t \ge k+1`.
    :type time: pd.DataFrame
    :param life_time: :math:`L^m_{l,o}(k)` is a `lifetime_mean_value` DataFrame at `time` :math:`k`.
    :type life_time: pd.DataFrame
    :param k: a `time` integer.
    :type k: int

    :return: :math:`d^e_{l,o,t}` a DataFrame resulting of the exponential decay function.
    :rtype: pd.DataFrame

    calculation
    -----------
    .. math::
        d^e_{l,o,t}[t,L^m_{l,o,t},k] = \frac{{e^{-\frac{{(t - k)}}{{L^m_{l,o}(k)}}}}}{{L^m_{l,o}(k)}}
        :label: exponential_decay

    Notes
    -----
    Indices are the :ref:`variables dimensions <dimensions>`.
    """
    return (
        np.exp(-(time.loc[:, k + 1 :] - k).div(life_time[k], axis=0))
        .div(life_time[k], axis=0)
        .dropna(axis=0, how="all")
    )


def profile(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self.profiler.enable()
        retval = func(self, *args, **kwargs)
        self.profiler.disable()
        return retval

    return wrapper


def _compute_pseudoinverse(matrix):
    """Compute the Moore-Penrose pseudoinverse of a given matrix.

    This function calculates the pseudoinverse of a matrix using the NumPy linear algebra library. The Moore-Penrose pseudoinverse is a generalization of the matrix inverse for square matrices and is applicable to non-square matrices as well.

    :param matrix: A pandas DataFrame representing the matrix for which the pseudoinverse is to be computed.
    :type matrix: pd.DataFrame
    :return: The pseudoinverse of the matrix as a NumPy array. This array has the same dimensions as the input matrix transposed, and it satisfies the four Moore-Penrose conditions.
    :rtype: np.ndarray

    The computation is performed using the NumPy library's `linalg.pinv` method, which is based on Singular Value Decomposition (SVD).

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> df = pd.DataFrame([[1, 2], [3, 4]])
    >>> compute_pseudoinverse(df)
    array([[-2. ,  1. ],
           [ 1.5, -0.5]])

    See Also
    --------
    `np.linalg.pinv <https://numpy.org/doc/stable/reference/generated/numpy.linalg.pinv.html>`_ : NumPy's method to compute the pseudoinverse of a matrix.

    References
    ----------
    - Penrose, R. (1955). A generalized inverse for matrices. Proceedings of the Cambridge Philosophical Society, 51, 406-413.
    - Ben-Israel, A., & Greville, T. N. E. (2003). Generalized inverses: Theory and applications. Springer Science & Business Media.

    """
    return np.linalg.pinv(matrix.values)


def _groupby_sum_empty(df: pd.DataFrame | pd.Series, group_index: list):
    if df.empty:
        # Create an empty MultiIndex with the desired names
        multi_index = pd.MultiIndex.from_tuples([], names=group_index)
        if isinstance(df, pd.DataFrame):
            # Return an empty DataFrame with expected index structure
            return pd.DataFrame([], index=multi_index, dtype=float)
        elif isinstance(df, pd.Series):
            # Return an empty Series with expected index structure
            return pd.Series([], index=multi_index, dtype=float)
        else:
            raise TypeError
    else:
        # Perform the groupby operation if the Series is not empty
        return df.groupby(level=group_index).sum()  # may be parameterized if we need something else than sum()


def _loc_dynamic(df: pd.DataFrame | pd.Series, level_slice: Dict[str, List[str]]):
    """
    Efficiently filter a MultiIndex DataFrame or Series based on specific level slices.
    Still it is slower than direct DataFrame.loc so use it only if needed

    Parameters:
    - df: A pandas DataFrame or Series with a MultiIndex.
    - level_slice: A dictionary where keys are MultiIndex level names and values are lists of desired elements.

    Returns:
    - A filtered DataFrame or Series.
    """
    # Prepare the index slices
    idx = pd.IndexSlice
    slice_obj = [slice(None)] * len(df.index.levels)  # Slice all levels initially

    # Optimize filtering for each level
    for key, filter_values in level_slice.items():
        # Find the position of the level
        level_id = df.index.names.index(key)

        # Use Index's native intersection (faster than converting to list and set)
        filtered_slice = df.index.get_level_values(key).intersection(filter_values)

        # Update the slice object with the filtered values
        slice_obj[level_id] = filtered_slice

    # Apply slicing dynamically based on DataFrame or Series type
    filtered_df = df.loc[idx[tuple(slice_obj)], :] if isinstance(df, pd.DataFrame) else df.loc[idx[tuple(slice_obj)]]

    return filtered_df


def _loc_dynamic_direct(df: pd.DataFrame | pd.Series, level_slice: Dict[str, List[str]]):
    """
    Filter a MultiIndex DataFrame or Series by constructing the loc[] argument directly.

    Parameters:
    - df: A pandas DataFrame or Series with a MultiIndex.
    - level_slice: A dictionary where keys are MultiIndex level names and values are lists of desired elements.

    Returns:
    - A filtered DataFrame or Series.
    """
    # Known level order (use the order of MultiIndex levels)
    level_names = df.index.names

    # Build the tuple directly for loc[] using comprehension
    loc_args = tuple(
        df.index.get_level_values(level).intersection(level_slice[level]) if level in level_slice else slice(None)
        for level in level_names
    )

    # Apply loc[] directly
    return df.loc[loc_args]

==> mater-0.10.0-py3-none-any.whl: mater/_utils.py <==
"""
This module creates the external funcions used in the MATER class.
"""

import datetime
import glob
import os
from functools import wraps
from typing import Dict, List, Literal

import numpy as np
import pandas as pd
from data_completion_tool import dct
from pandas.core.frame import DataFrame

# Functions definition


def _log_normal(
    life_time: DataFrame,
    standard_deviation: DataFrame,
    k: datetime.datetime,
    offset: datetime.datetime,
) -> DataFrame:
    r"""
    Compute the log-normal probability density function.

    This function calculates the log-normal distribution given the lifetime mean value, standard deviation, and a set of time points. It is typically used to model time until failure or time until an event occurs, where the event times are assumed to follow a log-normal distribution.

    :param time: :math:`t` a DataFrame containing time points at which the log-normal distribution is evaluated.
    :param life_time: :math:`\mu_{l,o,t}` a DataFrame containing the lifetime mean values.
    :param standard_deviation: :math:`\sigma_{l,o,t}` a DataFrame containing the standard deviations of the lifetime values.
    :param k: A `time` integer.
    :type time: pd.DataFrame
    :type life_time: pd.DataFrame
    :type standard_deviation: pd.DataFrame
    :type k: int
    :return: :math:`d^{log}_{l,o,t}` a DataFrame containing the computed log-normal probability densities at each time point.
    :rtype: pd.DataFrame

    Calculation
    -----------
    The probability density function for a log-normal distribution is calculated as follows:

    .. math::
        d^{log}_{l,o,t}[t,\mu_{l,o,t},\sigma_{l,o,t},k] = \frac{1}{(t - k) \sigma_{l,o,t} \sqrt{2\pi}} \exp\left(-\frac{(\log(t - k) - \mu_{l,o,t})^2}{2\sigma_{l,o,t}^2}\right)
        :label: probability_density_function

    Notes
    -----
    Indices are the :ref:`variables dimensions <dimensions>`.
    """
    time = _time_df(life_time, k, offset)
    log_normal = np.exp(
        -1
        / 2
        * (
            (np.log(time).subtract(life_time[k], axis=0)).div(
                standard_deviation[k], axis=0
            )
        )
        ** 2
    ) / (time.mul(standard_deviation[k], axis=0) * np.sqrt(2 * np.pi))
    return log_normal


def _exponential_decay(
    life_time: DataFrame, k: datetime.datetime, offset: datetime.datetime
):
    r"""
    Calculate the exponential decay for given time points and decay lifetimes.

    :param time: :math:`t` is a `Time` DataFrame fot `time` :math:`t \ge k+1`.
    :type time: pd.DataFrame
    :param life_time: :math:`L^m_{l,o}(k)` is a `lifetime_mean_value` DataFrame at `time` :math:`k`.
    :type life_time: pd.DataFrame
    :param k: a `time` integer.
    :type k: int

    :return: :math:`d^e_{l,o,t}` a DataFrame resulting of the exponential decay function.
    :rtype: pd.DataFrame

    calculation
    -----------
    .. math::
        d^e_{l,o,t}[t,L^m_{l,o,t},k] = \frac{{e^{-\frac{{(t - k)}}{{L^m_{l,o}(k)}}}}}{{L^m_{l,o}(k)}}
        :label: exponential_decay

    Notes
    -----
    Indices are the :ref:`variables dimensions <dimensions>`.
    """
    ### WARNING: the lifetime value in life_time must me the same unit as the time_frequency
    ### ToDo : Handle automatic conversion depending on the unit of life_time and the given time_frequency
    time = _time_df(life_time, k, offset)
    exponential_decay = (
        np.exp(-time.div(life_time[k], axis=0))
        .div(life_time[k], axis=0)
        .dropna(axis=0, how="all")
    )
    return exponential_decay


def _time_df(life_time: DataFrame, k: datetime.datetime, offset: datetime.datetime):
    col = life_time.loc[:, k + offset :].columns
    length = len(col)
    t = np.arange(1, length + 1, 1)
    time = pd.DataFrame(
        data=[t] * len(life_time.index), index=life_time.index, columns=col
    )
    time.columns.name = "time"
    return time


def profile(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self.profiler.enable()
        retval = func(self, *args, **kwargs)
        self.profiler.disable()
        return retval

    return wrapper


def _compute_pseudoinverse(matrix):
    """Compute the Moore-Penrose pseudoinverse of a given matrix.

    This function calculates the pseudoinverse of a matrix using the NumPy linear algebra library. The Moore-Penrose pseudoinverse is a generalization of the matrix inverse for square matrices and is applicable to non-square matrices as well.

    :param matrix: A pandas DataFrame representing the matrix for which the pseudoinverse is to be computed.
    :type matrix: pd.DataFrame
    :return: The pseudoinverse of the matrix as a NumPy array. This array has the same dimensions as the input matrix transposed, and it satisfies the four Moore-Penrose conditions.
    :rtype: np.ndarray

    The computation is performed using the NumPy library's `linalg.pinv` method, which is based on Singular Value Decomposition (SVD).

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> df = pd.DataFrame([[1, 2], [3, 4]])
    >>> compute_pseudoinverse(df)
    array([[-2. ,  1. ],
           [ 1.5, -0.5]])

    See Also
    --------
    `np.linalg.pinv <https://numpy.org/doc/stable/reference/generated/numpy.linalg.pinv.html>`_ : NumPy's method to compute the pseudoinverse of a matrix.

    References
    ----------
    - Penrose, R. (1955). A generalized inverse for matrices. Proceedings of the Cambridge Philosophical Society, 51, 406-413.
    - Ben-Israel, A., & Greville, T. N. E. (2003). Generalized inverses: Theory and applications. Springer Science & Business Media.

    """
    return np.linalg.pinv(matrix.values)


def _groupby_sum_empty(df: pd.DataFrame | pd.Series, group_index: list):
    if df.empty:
        # Create an empty MultiIndex with the desired names
        multi_index = pd.MultiIndex.from_tuples([], names=group_index)
        if isinstance(df, pd.DataFrame):
            # Return an empty DataFrame with expected index structure
            return pd.DataFrame([], index=multi_index, dtype=float)
        elif isinstance(df, pd.Series):
            # Return an empty Series with expected index structure
            return pd.Series([], index=multi_index, dtype=float)
        else:
            raise TypeError
    else:
        # Perform the groupby operation if the Series is not empty
        return df.groupby(
            level=group_index
        ).sum()  # may be parameterized if we need something else than sum()


def _loc_dynamic(df: pd.DataFrame | pd.Series, level_slice: Dict[str, List[str]]):
    """
    Efficiently filter a MultiIndex DataFrame or Series based on specific level slices.
    Still it is slower than direct DataFrame.loc so use it only if needed

    Parameters:
    - df: A pandas DataFrame or Series with a MultiIndex.
    - level_slice: A dictionary where keys are MultiIndex level names and values are lists of desired elements.

    Returns:
    - A filtered DataFrame or Series.
    """
    # Prepare the index slices
    idx = pd.IndexSlice
    slice_obj = [slice(None)] * len(df.index.levels)  # Slice all levels initially

    # Optimize filtering for each level
    for key, filter_values in level_slice.items():
        # Find the position of the level
        level_id = df.index.names.index(key)

        # Use Index's native intersection (faster than converting to list and set)
        filtered_slice = df.index.get_level_values(key).intersection(filter_values)

        # Update the slice object with the filtered values
        slice_obj[level_id] = filtered_slice

    # Apply slicing dynamically based on DataFrame or Series type
    filtered_df = (
        df.loc[idx[tuple(slice_obj)], :]
        if isinstance(df, pd.DataFrame)
        else df.loc[idx[tuple(slice_obj)]]
    )

    return filtered_df


def _loc_dynamic_direct(
    df: pd.DataFrame | pd.Series, level_slice: Dict[str, List[str]]
):
    """
    Filter a MultiIndex DataFrame or Series by constructing the loc[] argument directly.

    Parameters:
    - df: A pandas DataFrame or Series with a MultiIndex.
    - level_slice: A dictionary where keys are MultiIndex level names and values are lists of desired elements.

    Returns:
    - A filtered DataFrame or Series.
    """
    # Known level order (use the order of MultiIndex levels)
    level_names = df.index.names

    # Build the tuple directly for loc[] using comprehension
    loc_args = tuple(
        df.index.get_level_values(level).intersection(level_slice[level])
        if level in level_slice
        else slice(None)
        for level in level_names
    )

    # Apply loc[] directly
    return df.loc[loc_args]


def convert_to_datetime(column: pd.Series) -> pd.Series:
    """
    Converts a column of mixed date formats (float, string, or int) to pandas datetime.
    Handles formats like YYYYMMDD and epoch time systematically.

    :param column: pandas Series containing date values.
    :return: pandas Series with converted datetime values.
    """

    def parse_date(value):
        try:
            # Handle float and integer
            if isinstance(value, (float, int)):
                return pd.to_datetime(str(int(value)), errors="coerce")

            # Handle strings
            elif isinstance(value, str):
                return pd.to_datetime(value, errors="coerce")

        except Exception:
            return pd.NaT

    # Apply the parsing function to each value in the column
    return column.apply(parse_date)


def read_json(
    folder: Literal["input_data", "dimension", "variable_dimension"],
    scenario: str | None = None,
):
    folder_path = os.path.join("data", folder)
    pattern = os.path.join(folder_path, "*.json")
    json_files = glob.glob(pattern)
    dfs = [pd.read_json(f, orient="records") for f in json_files]
    df = pd.concat(dfs, ignore_index=True)
    if folder == "input_data":
        variable_names = df["variable"].unique()
        df = df[df["scenario"].isin(["historical", scenario])]
        dict_dfs = {
            var: df[df["variable"] == var].copy().drop(["variable", "scenario"], axis=1)
            for var in variable_names
        }
        return dict_dfs
    else:
        return df


def input_completion(
    input_name: str,
    input: pd.DataFrame,
    ds: dct.DataSet,
    variable_dimension: pd.DataFrame,
):
    input["time"] = convert_to_datetime(input["time"])
    variable_aspects = variable_dimension[variable_dimension["variable"] == input_name]
    aspect_property = (
        variable_aspects[["dimension", "property"]]
        .set_index("dimension")
        .T.to_dict("list")
    )
    completed_input = ds.completion(input, aspect_property)
    return completed_input
